
// To compile: g++ 2.cpp -o 2
// To run: ./2 [depth]
// (it just runs the test that's in main() with a tree of depth [depth] or 5,
// and wastes potentially a lot of time (which you can measure if you use
// 'time' or such), and finishes.)
//
// See the comment above breadthOrDepthFirstSearch() for complexity analysis.
// There is a main() function, but I did not do enough testing on this code
// (you should grade it less for that if good code is part of this grade).
//
// I first tried writing it in SML but I didn't know that language well enough
// to do that efficiently or its online docs weren't good enough (see
// rationale and attempt in 2.sml).


#include <stdlib.h>

#include <list>
#include <vector>
#include <algorithm>

// This is a horrid mash of C and C++.  I do not even bother to have code freeing
// the memory allocated for the tree.  I use comments to note C++ implicit
// behaviors.

enum CoinFlip { Heads, Tails };

// If we don't store any data in this tree, then the entire large
// tree might as well just be represented by one int (its depth)
// and if it were an abstract data type, no one could tell!
// Or by something similar to TreeWithHistory below.
struct CoinFlipTree {
	// These pointers are both null if at a leaf of the tree,
	// or both nonnull otherwise.
	CoinFlipTree* ifHeads;
	CoinFlipTree* ifTails;
};

CoinFlipTree* generateCoinFlipTree(unsigned int depth) {
	if(depth == 0) {
		CoinFlipTree* tree = (CoinFlipTree*)malloc(sizeof(CoinFlipTree));
		tree->ifHeads = NULL;
		tree->ifTails = NULL;
		return tree;
	}
	else {
		CoinFlipTree* tree = (CoinFlipTree*)malloc(sizeof(CoinFlipTree));
		tree->ifHeads = generateCoinFlipTree(depth - 1);
		tree->ifTails = generateCoinFlipTree(depth - 1);
		return tree;
	}
}

// This is actually a problem that we can search without thinking about
// a frontier at all, because there is only one correct path that we can
// find out rather trivially.
//
// There is no reason this uses recursion; if I was thinking like
// an imperative programmer, or was optimizing C/C++ performance, or was
// worried about large arrays causing stack overflow, I would make it
// a loop instead.
CoinFlipTree* obviousSearch(CoinFlipTree* tree,
		unsigned int numCoinFlips, CoinFlip* coinFlipsArray) {
	if(numCoinFlips == 0) {
		return tree;
	}
	else if(tree == NULL) {
		return NULL;
	}
	else {
		CoinFlipTree* subtree;
		if(coinFlipsArray[0] == Heads) {
			subtree = tree->ifHeads;
		}
		else {
			subtree = tree->ifTails;
		}
		return obviousSearch(subtree, numCoinFlips - 1, coinFlipsArray + 1);
	}
}

// It's easier to think about all the state implied by these algorithms if
// we eschew recursion.  I'm using C++ STL libraries for this internal state
// because I know them and explicit memory management would obfuscate this
// code.

//
struct TreeWithHistory {
	CoinFlipTree* tree;
	std::vector<CoinFlip> history; //implicitly inited to []
};

// Because C++ is too sad to be able to have even vector deep equality
// built in.
bool history_equal(std::vector<CoinFlip> const& history,
		unsigned int numCoinFlips, CoinFlip* coinFlipsArray) {
	if(numCoinFlips != history.size()) {
		return false;
	}
	return std::equal(history.begin(), history.end(), coinFlipsArray);
}
bool history_is_a_prefix(std::vector<CoinFlip> const& history,
		unsigned int numCoinFlips, CoinFlip* coinFlipsArray) {
	if(numCoinFlips < history.size()) {
		return false;
	}
	return std::equal(history.begin(), history.end(), coinFlipsArray);
}

enum BreadthOrDepthFirst {
	BreadthFirst, DepthFirst
};

// stupid == true: a pure breadth first or depth first search.
// If n is the number of coin flips, both breadth and depth first will
// search O(2^n) nodes on average, although depth-first has an unlikely
// best-case time of O(n).  Because of how this search checks whether it's
// found the right node, average case time is O(n*2^n).
// (If n represents the number of items in the tree, then the average case
// time is O(n) or rather O(n log n).)
//
// stupid == false:
// We prune nodes from being searched farther if we can tell they're
// wrong by the time we see them.
// If n is the number of coin flips, both breadth and depth first will
// search O(n) nodes before finding the right answer (and then check that
// it is the right answer at mere O(n) cost).  If n is the number
// of items in the tree, they will search O(log n) nodes.
CoinFlipTree* breadthOrDepthFirstSearch(
		CoinFlipTree* tree,
		BreadthOrDepthFirst searchOrder,
		bool stupid,
		unsigned int numCoinFlips, CoinFlip* coinFlipsArray) {
       	// implicitly inited to [];
	// used as a FIFO (breadth-first) or LIFO (depth-first) queue.
	std::list<TreeWithHistory> frontier;
	TreeWithHistory initial;
	initial.tree = tree;
	frontier.push_back(initial);

	TreeWithHistory current = initial;
	while(!history_equal(current.history, numCoinFlips, coinFlipsArray)) {
		if(stupid || history_is_a_prefix(current.history,
		                                 numCoinFlips, coinFlipsArray)) {
			if(current.tree != NULL) {
				TreeWithHistory next1 = current; //deep copies 'history'
				next1.tree = current.tree->ifHeads;
				next1.history.push_back(Heads);
				frontier.push_back(next1);

				TreeWithHistory next2 = current; //deep copies 'history'
				next2.tree = current.tree->ifTails;
				next2.history.push_back(Tails);
				frontier.push_back(next2);
			}
		}
		if(frontier.empty()) {
			return NULL;
		}
		if(searchOrder == BreadthFirst) {
			current = frontier.front();
			frontier.pop_front();
		}
		else {
			current = frontier.back();
			frontier.pop_back();
		}
	}
	return current.tree;
}


int main(int argc, char* argv[]) {
	long depth = 5;
	if(argc >= 2) {
		depth = strtol(argv[1], NULL, 10);
	}
	CoinFlipTree* tree = generateCoinFlipTree(depth);
	std::vector<CoinFlip> coinFlipsArrayToSearchFor;
	for(int i = 0; i != depth; ++i) {
		// check some arbitrary bit of rand()
		if(rand()&(1<<13)) {
			coinFlipsArrayToSearchFor.push_back(Heads);
		}
		else {
			coinFlipsArrayToSearchFor.push_back(Tails);
		}
	}
	BreadthOrDepthFirst searchOrder = BreadthFirst;
	bool stupid = true;
	breadthOrDepthFirstSearch(tree, searchOrder, stupid, depth,
	                          &coinFlipsArrayToSearchFor[0]);
	return 0;
}
