
-- This is a fairly straightforward adaptation of 2B.js.
-- I looked for a CSV library to show how great it is to use libs
-- (in any language) but the one I saw didn't support arbitrary
-- separator characters (i.e. semicolons).

-- After writing it, I was bemused by how many fewer lines the
-- parsing functions were, even though they are saying the same
-- thing.  I think it's a combination of currying, a shorter
-- lambda than function(){return}, and in parseFamilyTreeIntoRows,
-- pattern matching.
-- A line in transformRowsToGraphvizDotText shows that I wrap lines
-- a bit less in Haskell, a combination of style habits and how
-- Haskell's libs are slightly more concise (more functions and
-- they don't have _. at the beginning of them).
--
-- Also, it ran on the first try after fixing type errors.
-- And it was fun to write.  Blargh I am too good at this.
       
-- usage:
--   cat 2B.csv | runghc 2B.hs | dot -Tpng | display -

import qualified Data.Set as Set
import qualified Data.Map as Map

split :: Char -> String -> [String]
split sep = go []
  where
    go :: String -> String -> [String]
    go accum [] = [reverse accum]
    go accum (c:cs)
      | c == sep  = reverse accum : go [] cs
      | otherwise = go (c:accum) cs

parseCsvText :: String -> [[String]]
parseCsvText = map (split ';') . filter (/= "") . split '\n'

data Row = Row {
  rowID :: String,
  rowName :: String,
  rowParents :: [String],
  rowBirthdate :: String,
  rowDeathdate :: String
}
parseFamilyTreeIntoRows :: String -> [Row]
parseFamilyTreeIntoRows text =
  map mkrow (parseCsvText text)
  where
    mkrow [rid, name, parents, birthdate, deathdate] =
      Row rid name (split ',' parents) birthdate deathdate

transformRowsToGraphvizDotText :: [Row] -> String
transformRowsToGraphvizDotText rows = let
    describedPeopleIDs = Set.fromList (map rowID rows)
    mentionedPeopleIDs = Set.fromList (map rowID rows ++ concatMap rowParents rows)
    rowsByID = Map.fromList (zip (map rowID rows) rows)
    showIDInfo personID = case Map.lookup personID rowsByID of
      Nothing -> ["\""++personID++"\" [label=\"??\"]"]
      Just row -> ["\""++personID++"\" [label=\""++rowName row++"\"]"] ++
                  map (\parentID -> "\""++parentID++"\" -> \""++personID++"\"")
                      (rowParents row)
  in
    unlines (
      ["digraph \"family tree\" {"] ++
      concatMap showIDInfo (Set.toList mentionedPeopleIDs)
        ++ ["}"]
    )

main :: IO ()
main = do
  csvText <- getContents
  putStr (transformRowsToGraphvizDotText (parseFamilyTreeIntoRows csvText))
