#!/usr/bin/env node

// usage:
//   (edit 2B.csv)
//   node 2B.js | dot -Tpng | display -
// I wanted:
//   cat 2B.csv | node 2B.js | dot -Tpng | display -
// but node was giving me issues with importing process.stdin.
//
// This file could easily call all of these things
// in the pipeline itself, but it seems more Unixy to do it this way.

var _ = require('underscore');
var fs = require('fs');
var csvText = fs.readFileSync('2B.csv', 'utf8');

function parseCsvText(text) {
  var lines = _.filter(text.split('\n'), function(line) {
    return line.length > 0;
  });
  return _.map(lines, function(line) {
    var fields = line.split(';');
    return fields;
  });
}

function parseFamilyTreeIntoRows(text) {
  var lines = parseCsvText(text);
  var rows = [];
  _.each(lines, function(line) {
    rows.push({
      id: line[0],
      name: line[1],
      parents: line[2].split(','),
      birthdate: line[3],
      deathdate: line[4]
    });
  });
  return rows;
}

function asHashSet(arr) {
  var result = {};
  _.each(arr, function(elem) {
    result[elem] = true;
  });
  return result;
}

function transformRowsToGraphvizDotText(rows) {
  var describedPeopleIDs = asHashSet(_.pluck(rows, 'id'));
  var mentionedPeopleIDs = asHashSet(
                          _.flatten(_.pluck(rows, 'parents')).
                             concat(_.pluck(rows, 'id'))
                           );
  //var rowsByID = _.object(_.map(rows, function(row) { return [row.id, row]; }));
  var rowsByID = _.object(_.pluck(rows, 'id'), rows);

  var resultLines = [];
  function line(l) {
    resultLines.push(l + '\n');
  }
  line('digraph "family tree" {');
  _.each(_.keys(mentionedPeopleIDs).sort(), function(id) {
    var row = rowsByID[id];
    if(row === undefined) {
      line('"'+id+'" [label="??"]');
    }
    else {
      line('"'+id+'" [label="'+row.name+'"]');
      _.each(row.parents, function(parent) {
        line('"'+parent+'" -> "'+id+'"');
      });
    }
  });
  line('}');
  var result = ''.concat.apply('', resultLines);
  return result;
}

console.log(transformRowsToGraphvizDotText(parseFamilyTreeIntoRows(csvText)));
